# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import os
import pickle
from random import randint
from uuid import uuid4

import pytest


@pytest.mark.need_capability("read", "write", "delete", "shared")
def test_operator_pickle(service_name, operator, async_operator):
    """
    Test Operator's pickle serialization and deserialization.
    """

    size = randint(1, 1024)
    filename = f"random_file_{str(uuid4())}"
    content = os.urandom(size)
    operator.write(filename, content)

    serialized = pickle.dumps(operator)

    deserialized = pickle.loads(serialized)
    assert deserialized.read(filename) == content

    operator.delete(filename)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import pickle
from datetime import datetime
from uuid import uuid4

import pytest


@pytest.mark.need_capability("read", "write", "delete")
def test_sync_file_pickle(service_name, operator, async_operator):
    """
    Test pickle streaming serialization and deserialization using OpenDAL operator.
    """
    data = {
        "a": 1,
        "b": "hello",
        "c": [1, 2, 3],
        "d": {"e": 4},
        "f": None,
        "g": b"hello\nworld",
        "h": set([1, 2, 3]),
        "i": 1.23,
        "j": True,
        "k": datetime.strptime("2024-01-01", "%Y-%m-%d"),
    }

    filename = f"random_file_{str(uuid4())}"
    with operator.open(filename, "wb") as f:
        pickle.dump(data, f)

    with operator.open(filename, "rb") as f:
        assert pickle.load(f) == data



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

class Error(Exception):
    """Base class for exceptions in this module."""

class Unexpected(Error):
    """Unexpected errors"""

class Unsupported(Error):
    """Unsupported operation"""

class ConfigInvalid(Error):
    """Config is invalid"""

class NotFound(Error):
    """Not found"""

class PermissionDenied(Error):
    """Permission denied"""

class IsADirectory(Error):
    """Is a directory"""

class NotADirectory(Error):
    """Not a directory"""

class AlreadyExists(Error):
    """Already exists"""

class IsSameFile(Error):
    """Is same file"""

class ConditionNotMatch(Error):
    """Condition not match"""



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
> DO NOT EDIT IT MANUALLY <

This file is generated by opendal/dev/src/generate/python.rs.
`opendal.__base` doesn't exists.
"""

from typing import Literal, TypeAlias, overload

# `true`/`false`` in any case, for example, `true`/`True`/`TRUE` `false`/`False`/`FALSE`
_bool: TypeAlias = str
# a str represent a int, for example, `"10"`/`"0"`
_int: TypeAlias = str

# a human readable duration string
# see https://docs.rs/humantime/latest/humantime/fn.parse_duration.html
# for more details
_duration: TypeAlias = str


# A "," separated string, for example `"127.0.0.1:1,127.0.0.1:2"`
_strings: TypeAlias = str

class _Base:
    """this is not a real base class but typing mixin,

    The services list here is support by opendal pypi wheel.
    """


    @overload
    def __init__(
        self,
        scheme: Literal["aliyun_drive"],
        /,*,
        drive_type: str,
        root: str = ...,
        access_token: str = ...,
        client_id: str = ...,
        client_secret: str = ...,
        refresh_token: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["alluxio"],
        /,*,
        root: str = ...,
        endpoint: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["atomicserver"],
        /,*,
        root: str = ...,
        endpoint: str = ...,
        private_key: str = ...,
        public_key: str = ...,
        parent_resource_id: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["azblob"],
        /,*,
        container: str,
        root: str = ...,
        endpoint: str = ...,
        account_name: str = ...,
        account_key: str = ...,
        encryption_key: str = ...,
        encryption_key_sha256: str = ...,
        encryption_algorithm: str = ...,
        sas_token: str = ...,
        batch_max_operations: _int = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["azdls"],
        /,*,
        filesystem: str,
        root: str = ...,
        endpoint: str = ...,
        account_name: str = ...,
        account_key: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["azfile"],
        /,*,
        share_name: str,
        root: str = ...,
        endpoint: str = ...,
        account_name: str = ...,
        account_key: str = ...,
        sas_token: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["b2"],
        /,*,
        bucket: str,
        bucket_id: str,
        root: str = ...,
        application_key_id: str = ...,
        application_key: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["cacache"],
        /,*,
        datadir: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["cloudflare_kv"],
        /,*,
        token: str = ...,
        account_id: str = ...,
        namespace_id: str = ...,
        root: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["compfs"],
        /,*,
        root: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["cos"],
        /,*,
        root: str = ...,
        endpoint: str = ...,
        secret_id: str = ...,
        secret_key: str = ...,
        bucket: str = ...,
        enable_versioning: _bool = ...,
        disable_config_load: _bool = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["d1"],
        /,*,
        token: str = ...,
        account_id: str = ...,
        database_id: str = ...,
        root: str = ...,
        table: str = ...,
        key_field: str = ...,
        value_field: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["dashmap"],
        /,*,
        root: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["dbfs"],
        /,*,
        root: str = ...,
        endpoint: str = ...,
        token: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["dropbox"],
        /,*,
        root: str = ...,
        access_token: str = ...,
        refresh_token: str = ...,
        client_id: str = ...,
        client_secret: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["fs"],
        /,*,
        root: str = ...,
        atomic_write_dir: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["gcs"],
        /,*,
        bucket: str,
        root: str = ...,
        endpoint: str = ...,
        scope: str = ...,
        service_account: str = ...,
        credential: str = ...,
        credential_path: str = ...,
        predefined_acl: str = ...,
        default_storage_class: str = ...,
        allow_anonymous: _bool = ...,
        disable_vm_metadata: _bool = ...,
        disable_config_load: _bool = ...,
        token: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["gdrive"],
        /,*,
        root: str = ...,
        access_token: str = ...,
        refresh_token: str = ...,
        client_id: str = ...,
        client_secret: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["ghac"],
        /,*,
        root: str = ...,
        version: str = ...,
        endpoint: str = ...,
        runtime_token: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["github"],
        /,*,
        owner: str,
        repo: str,
        root: str = ...,
        token: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["gridfs"],
        /,*,
        connection_string: str = ...,
        database: str = ...,
        bucket: str = ...,
        chunk_size: _int = ...,
        root: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["hdfs_native"],
        /,*,
        root: str = ...,
        name_node: str = ...,
        enable_append: _bool = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["http"],
        /,*,
        endpoint: str = ...,
        username: str = ...,
        password: str = ...,
        token: str = ...,
        root: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["huggingface"],
        /,*,
        repo_type: str = ...,
        repo_id: str = ...,
        revision: str = ...,
        root: str = ...,
        token: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["icloud"],
        /,*,
        root: str = ...,
        apple_id: str = ...,
        password: str = ...,
        trust_token: str = ...,
        ds_web_auth_token: str = ...,
        is_china_mainland: _bool = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["ipfs"],
        /,*,
        endpoint: str = ...,
        root: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["ipmfs"],
        /,*,
        root: str = ...,
        endpoint: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["koofr"],
        /,*,
        endpoint: str,
        email: str,
        root: str = ...,
        password: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["lakefs"],
        /,*,
        endpoint: str = ...,
        username: str = ...,
        password: str = ...,
        root: str = ...,
        repository: str = ...,
        branch: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["memcached"],
        /,*,
        endpoint: str = ...,
        root: str = ...,
        username: str = ...,
        password: str = ...,
        default_ttl: _duration = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["memory"],
        /,*,
        root: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["mini_moka"],
        /,*,
        max_capacity: _int = ...,
        time_to_live: _duration = ...,
        time_to_idle: _duration = ...,
        root: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["moka"],
        /,*,
        name: str = ...,
        max_capacity: _int = ...,
        time_to_live: _duration = ...,
        time_to_idle: _duration = ...,
        num_segments: _int = ...,
        root: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["mongodb"],
        /,*,
        connection_string: str = ...,
        database: str = ...,
        collection: str = ...,
        root: str = ...,
        key_field: str = ...,
        value_field: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["monoiofs"],
        /,*,
        root: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["mysql"],
        /,*,
        connection_string: str = ...,
        table: str = ...,
        key_field: str = ...,
        value_field: str = ...,
        root: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["nebula_graph"],
        /,*,
        host: str = ...,
        port: _int = ...,
        username: str = ...,
        password: str = ...,
        space: str = ...,
        tag: str = ...,
        key_field: str = ...,
        value_field: str = ...,
        root: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["obs"],
        /,*,
        root: str = ...,
        endpoint: str = ...,
        access_key_id: str = ...,
        secret_access_key: str = ...,
        bucket: str = ...,
        enable_versioning: _bool = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["onedrive"],
        /,*,
        root: str = ...,
        access_token: str = ...,
        refresh_token: str = ...,
        client_id: str = ...,
        client_secret: str = ...,
        enable_versioning: _bool = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["opfs"],
        /,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["oss"],
        /,*,
        bucket: str,
        root: str = ...,
        endpoint: str = ...,
        presign_endpoint: str = ...,
        enable_versioning: _bool = ...,
        server_side_encryption: str = ...,
        server_side_encryption_key_id: str = ...,
        allow_anonymous: _bool = ...,
        access_key_id: str = ...,
        access_key_secret: str = ...,
        # deprecated: Please use `delete_max_size` instead of `batch_max_operations`
        batch_max_operations: _int = ...,
        delete_max_size: _int = ...,
        role_arn: str = ...,
        role_session_name: str = ...,
        oidc_provider_arn: str = ...,
        oidc_token_file: str = ...,
        sts_endpoint: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["pcloud"],
        /,*,
        endpoint: str,
        root: str = ...,
        username: str = ...,
        password: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["persy"],
        /,*,
        datafile: str = ...,
        segment: str = ...,
        index: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["postgresql"],
        /,*,
        root: str = ...,
        connection_string: str = ...,
        table: str = ...,
        key_field: str = ...,
        value_field: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["redb"],
        /,*,
        datadir: str = ...,
        root: str = ...,
        table: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["redis"],
        /,*,
        db: _int,
        endpoint: str = ...,
        cluster_endpoints: str = ...,
        username: str = ...,
        password: str = ...,
        root: str = ...,
        default_ttl: _duration = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["s3"],
        /,*,
        bucket: str,
        root: str = ...,
        enable_versioning: _bool = ...,
        endpoint: str = ...,
        region: str = ...,
        access_key_id: str = ...,
        secret_access_key: str = ...,
        session_token: str = ...,
        role_arn: str = ...,
        external_id: str = ...,
        role_session_name: str = ...,
        disable_config_load: _bool = ...,
        disable_ec2_metadata: _bool = ...,
        allow_anonymous: _bool = ...,
        server_side_encryption: str = ...,
        server_side_encryption_aws_kms_key_id: str = ...,
        server_side_encryption_customer_algorithm: str = ...,
        server_side_encryption_customer_key: str = ...,
        server_side_encryption_customer_key_md5: str = ...,
        default_storage_class: str = ...,
        enable_virtual_host_style: _bool = ...,
        # deprecated: Please use `delete_max_size` instead of `batch_max_operations`
        batch_max_operations: _int = ...,
        delete_max_size: _int = ...,
        disable_stat_with_override: _bool = ...,
        checksum_algorithm: str = ...,
        disable_write_with_if_match: _bool = ...,
        enable_write_with_append: _bool = ...,
        disable_list_objects_v2: _bool = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["seafile"],
        /,*,
        repo_name: str,
        root: str = ...,
        endpoint: str = ...,
        username: str = ...,
        password: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["sftp"],
        /,*,
        endpoint: str = ...,
        root: str = ...,
        user: str = ...,
        key: str = ...,
        known_hosts_strategy: str = ...,
        enable_copy: _bool = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["sled"],
        /,*,
        datadir: str = ...,
        root: str = ...,
        tree: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["sqlite"],
        /,*,
        connection_string: str = ...,
        table: str = ...,
        key_field: str = ...,
        value_field: str = ...,
        root: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["surrealdb"],
        /,*,
        connection_string: str = ...,
        username: str = ...,
        password: str = ...,
        namespace: str = ...,
        database: str = ...,
        table: str = ...,
        key_field: str = ...,
        value_field: str = ...,
        root: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["swift"],
        /,*,
        endpoint: str = ...,
        container: str = ...,
        root: str = ...,
        token: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["upyun"],
        /,*,
        bucket: str,
        root: str = ...,
        operator: str = ...,
        password: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["vercel_artifacts"],
        /,*,
        access_token: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["vercel_blob"],
        /,*,
        root: str = ...,
        token: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["webdav"],
        /,*,
        endpoint: str = ...,
        username: str = ...,
        password: str = ...,
        token: str = ...,
        root: str = ...,
        disable_copy: _bool = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["webhdfs"],
        /,*,
        root: str = ...,
        endpoint: str = ...,
        user_name: str = ...,
        delegation: str = ...,
        disable_list_batch: _bool = ...,
        atomic_write_dir: str = ...,
    ) -> None: ...

    @overload
    def __init__(
        self,
        scheme: Literal["yandex_disk"],
        /,*,
        access_token: str,
        root: str = ...,
    ) -> None: ...


    @overload
    def __init__(self, scheme: str, /, **kwargs: str) -> None: ...


# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import asyncio
from uuid import uuid4

import pytest


@pytest.mark.asyncio
@pytest.mark.need_capability("read", "write", "copy", "list", "list_with_start_after")
async def test_async_list_with_start_after(service_name, operator, async_operator):
    test_dir = f"test_async_list_dir_{uuid4()}/"
    await async_operator.create_dir(test_dir)

    # 1. Prepare data
    files_to_create = [f"{test_dir}file_{i}" for i in range(5)]
    # Create files concurrently
    await asyncio.gather(
        *(async_operator.write(f, b"test_content") for f in files_to_create)
    )

    # 2. Test basic list
    entries = []
    async for entry in await async_operator.list(test_dir):
        entries.append(entry.path)
    entries.sort()  # Ensure order for comparison
    expected_files = sorted([test_dir, *files_to_create])
    assert entries == expected_files, (
        f"Basic list failed. Expected {expected_files}, got {entries}"
    )

    # 3. Test list with start_after
    start_after_file = files_to_create[2]  # e.g., test_dir/file_2
    entries_after = []
    # Note: start_after expects the *full path* relative to the operator root
    async for entry in await async_operator.list(
        test_dir, start_after=start_after_file
    ):
        entries_after.append(entry.path)
    entries_after.sort()  # Ensure order

    # Expected files are those lexicographically after start_after_file
    expected_files_after = sorted([f for f in files_to_create if f > start_after_file])
    assert entries_after == expected_files_after, (
        f"Expected {expected_files_after} after {start_after_file}, got {entries_after}"
    )
    # 6. Cleanup
    await async_operator.remove_all(test_dir)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import os
from uuid import uuid4

import pytest

from opendal.exceptions import IsADirectory, IsSameFile, NotFound


@pytest.mark.asyncio
@pytest.mark.need_capability("read", "write", "copy")
async def test_async_copy(service_name, operator, async_operator):
    source_path = f"random_file_{str(uuid4())}"
    content = os.urandom(1024)
    await async_operator.write(source_path, content)
    target_path = f"random_file_{str(uuid4())}"
    await async_operator.copy(source_path, target_path)
    read_content = await async_operator.read(target_path)
    assert read_content is not None
    assert read_content == content
    await async_operator.delete(source_path)
    await async_operator.delete(target_path)


@pytest.mark.asyncio
@pytest.mark.need_capability("read", "write", "copy")
async def test_async_copy_non_exist(service_name, operator, async_operator):
    source_path = f"random_file_{str(uuid4())}"
    target_path = f"random_file_{str(uuid4())}"
    with pytest.raises(NotFound):
        await async_operator.copy(source_path, target_path)


@pytest.mark.asyncio
@pytest.mark.need_capability("read", "write", "copy", "create_dir")
async def test_async_copy_source_directory(service_name, operator, async_operator):
    source_path = f"random_file_{str(uuid4())}/"
    await async_operator.create_dir(source_path)
    target_path = f"random_file_{str(uuid4())}"
    with pytest.raises(IsADirectory):
        await async_operator.copy(source_path, target_path)


@pytest.mark.asyncio
@pytest.mark.need_capability("read", "write", "copy", "create_dir")
async def test_async_copy_target_directory(service_name, operator, async_operator):
    source_path = f"random_file_{str(uuid4())}"
    content = os.urandom(1024)
    await async_operator.write(source_path, content)
    target_path = f"random_file_{str(uuid4())}/"
    await async_operator.create_dir(target_path)
    with pytest.raises(IsADirectory):
        await async_operator.copy(source_path, target_path)
    await async_operator.delete(source_path)
    await async_operator.delete(target_path)


@pytest.mark.asyncio
@pytest.mark.need_capability("read", "write", "copy")
async def test_async_copy_self(service_name, operator, async_operator):
    source_path = f"random_file_{str(uuid4())}"
    content = os.urandom(1024)
    await async_operator.write(source_path, content)
    with pytest.raises(IsSameFile):
        await async_operator.copy(source_path, source_path)
    await async_operator.delete(source_path)


@pytest.mark.asyncio
@pytest.mark.need_capability("read", "write", "copy")
async def test_async_copy_nested(service_name, operator, async_operator):
    source_path = f"random_file_{str(uuid4())}"
    target_path = f"random_file_{str(uuid4())}/{str(uuid4())}/{str(uuid4())}"
    content = os.urandom(1024)
    await async_operator.write(source_path, content)
    await async_operator.copy(source_path, target_path)
    target_content = await async_operator.read(target_path)
    assert target_content is not None
    assert target_content == content
    await async_operator.delete(source_path)
    await async_operator.delete(target_path)


@pytest.mark.asyncio
@pytest.mark.need_capability("read", "write", "copy")
async def test_async_copy_overwrite(service_name, operator, async_operator):
    source_path = f"random_file_{str(uuid4())}"
    target_path = f"random_file_{str(uuid4())}"
    source_content = os.urandom(1024)
    target_content = os.urandom(1024)
    assert source_content != target_content
    await async_operator.write(source_path, source_content)
    await async_operator.write(target_path, target_content)
    await async_operator.copy(source_path, target_path)
    target_content = await async_operator.read(target_path)
    assert target_content is not None
    assert target_content == source_content
    await async_operator.delete(source_path)
    await async_operator.delete(target_path)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import os
from uuid import uuid4

import pytest
from dotenv import load_dotenv

import opendal

load_dotenv()
pytest_plugins = ("pytest_asyncio",)


def pytest_configure(config):
    # register an additional marker
    config.addinivalue_line(
        "markers",
        "need_capability(*capability): mark test to run only on named capability",
    )


@pytest.fixture(scope="session")
def service_name():
    service_name = os.environ.get("OPENDAL_TEST")
    if service_name is None:
        pytest.skip("OPENDAL_TEST not set")
    return service_name


@pytest.fixture(scope="session")
def setup_config(service_name):
    # Read arguments from envs.
    prefix = f"opendal_{service_name}_"
    config = {}
    for key in os.environ.keys():
        if key.lower().startswith(prefix):
            config[key[len(prefix) :].lower()] = os.environ.get(key)
    disable_random_root = (
        True if os.environ.get("OPENDAL_DISABLE_RANDOM_ROOT") == "true" else False
    )
    if not disable_random_root:
        config["root"] = f"{config.get('root', '/')}/{str(uuid4())}/"
    return config


@pytest.fixture(scope="session")
def async_operator(service_name, setup_config):
    return (
        opendal.AsyncOperator(service_name, **setup_config)
        .layer(opendal.layers.RetryLayer())
        .layer(opendal.layers.ConcurrentLimitLayer(1024))
        .layer(opendal.layers.MimeGuessLayer())
    )


@pytest.fixture(scope="session")
def operator(async_operator):
    return async_operator.to_operator()


@pytest.fixture(autouse=True)
def check_capability(request, operator, async_operator):
    if request.node.get_closest_marker("need_capability"):
        if request.node.get_closest_marker("need_capability").args:
            if not all(
                [
                    getattr(operator.capability(), x)
                    for x in request.node.get_closest_marker("need_capability").args
                ]
                + [
                    getattr(async_operator.capability(), x)
                    for x in request.node.get_closest_marker("need_capability").args
                ]
            ):
                pytest.skip(
                    "skip because "
                    f"{request.node.get_closest_marker('need_capability').args}"
                    " not supported"
                )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import asyncio
import timeit

from pydantic_settings import BaseSettings

import opendal


class Config(BaseSettings):
    aws_endpoint: str
    aws_region: str
    aws_s3_bucket: str


SETTINGS = Config()

TEST_CASE = [
    {"name": "4kb", "data": b"0" * 4096},
    {"name": "8kb", "data": b"0" * 8192},
    {"name": "16kb", "data": b"0" * 16384},
    {"name": "32kb", "data": b"0" * 32768},
    {"name": "256kb", "data": b"0" * 262144},
    {"name": "512kb", "data": b"0" * 524288},
    {"name": "16mb", "data": b"0" * 16777216},
]


async def opendal_write():
    op = opendal.AsyncOperator(
        "s3",
        bucket=SETTINGS.aws_s3_bucket,
        region=SETTINGS.aws_region,
        endpoint=SETTINGS.aws_endpoint,
    )
    tasks = []
    for case in TEST_CASE:
        tasks.append(
            op.write(
                f"/benchmark/opendal_write/{case['name']}",
                case["data"],
            )
        )
    await asyncio.gather(*tasks)


async def opendal_read():
    op = opendal.AsyncOperator(
        "s3",
        bucket=SETTINGS.aws_s3_bucket,
        region=SETTINGS.aws_region,
        endpoint=SETTINGS.aws_endpoint,
    )
    tasks = []
    for case in TEST_CASE:
        tasks.append(op.read(f"/benchmark/opendal_write/{case['name']}"))
    await asyncio.gather(*tasks)


def read_run():
    asyncio.run(opendal_read())


def write_run():
    asyncio.run(opendal_write())


def opendal_benchmark():
    print(f"OpenDAL S3 Client async write: {timeit.timeit(write_run, number=3)}")
    print(f"OpenDAL S3 Client async read: {timeit.timeit(read_run, number=3)}")


if __name__ == "__main__":
    opendal_benchmark()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import os
from types import TracebackType
from typing import Any, AsyncIterable, Iterable, Optional, Type, Union, final

from opendal import exceptions as exceptions
from opendal import layers as layers
from opendal.__base import _Base
from opendal.layers import Layer

PathBuf = Union[str, os.PathLike]

@final
class Operator(_Base):
    """The entry class for all public blocking APIs.

    Args:
        scheme (str): The service name that OpenDAL supports.
        **options (any): The options for the service.
            See the documentation of each service for more details.

    Example:
        ```python
        import opendal
        op = opendal.Operator("s3", bucket="bucket", region="us-east-1")
        op.write("hello.txt", b"hello world")
        ```
    """
    def __init__(self, scheme: str, **options: Any) -> None: ...
    def layer(self, layer: Layer) -> "Operator":
        """Add new layers upon the current operator.

        Args:
            layer (Layer): The layer to be added.

        Returns:
            The new operator with the layer added.
        """
    def open(self, path: PathBuf, mode: str) -> File:
        """Open a file at the given path for reading or writing.

        Args:
            path (str|Path): The path to the file.
            mode (str): The mode to open the file. Can be "rb" or "wb".

        Returns:
            A file-like object that can be used to read or write the file.

        Example:
            ```python
            import opendal
            op = opendal.Operator("s3", bucket="bucket", region="us-east-1")
            with op.open("hello.txt", "wb") as f:
                f.write(b"hello world")
            ```
        """
    def read(self, path: PathBuf) -> bytes:
        """Read the content of the object at the given path.

        Args:
            path (str|Path): The path to the object.

        Returns:
            The content of the object as bytes.
        """
    def write(
        self,
        path: PathBuf,
        bs: bytes,
        *,
        append: bool = ...,
        chunk: int = ...,
        content_type: str = ...,
        content_disposition: str = ...,
        cache_control: str = ...,
    ) -> None:
        """Write the content to the object at the given path.

        Args:
            path (str|Path): The path to the object.
            bs (bytes): The content to write.
            append (bool): Whether to append the content to the object.
                Defaults to False.
            chunk (int): The chunk size for writing. Defaults to write all.
            content_type (str): The content type of the object.
                Defaults to None.
            content_disposition (str): The content disposition of the object.
                Defaults to None.
            cache_control (str): The cache control of the object.
                Defaults to None.
        """
    def stat(self, path: PathBuf) -> Metadata:
        """Get the metadata of the object at the given path.

        Args:
            path (str|Path): The path to the object.

        Returns:
            The metadata of the object.
        """
    def create_dir(self, path: PathBuf) -> None:
        """Create a directory at the given path.

        Args:
            path (str|Path): The path to the directory.
        """
    def delete(self, path: PathBuf) -> None:
        """Delete the object at the given path.

        Args:
            path (str|Path): The path to the object.
        """
    def exists(self, path: PathBuf) -> bool:
        """Check if the object at the given path exists.

        Args:
            path (str|Path): The path to the object.

        Returns:
            True if the object exists, False otherwise.
        """
    def list(self, path: PathBuf, *, start_after: str | None = None) -> Iterable[Entry]:
        """List the objects at the given path.

        Args:
            path (str|Path): The path to the directory.
            start_after (str | None): The key to start listing from.

        Returns:
            An iterable of entries representing the objects in the directory.
        """
    def scan(self, path: PathBuf) -> Iterable[Entry]:
        """Scan the objects at the given path recursively.

        Args:
            path (str|Path): The path to the directory.

        Returns:
            An iterable of entries representing the objects in the directory.
        """
    def capability(self) -> Capability:
        """Get the capability of the operator.

        Returns:
            The capability of the operator.
        """
    def copy(self, source: PathBuf, target: PathBuf) -> None:
        """Copy the object from source to target.

        Args:
            source (str|Path): The source path.
            target (str|Path): The target path.
        """
    def rename(self, source: PathBuf, target: PathBuf) -> None:
        """Rename the object from source to target.

        Args:
            source (str|Path): The source path.
            target (str|Path): The target path.
        """
    def remove_all(self, path: PathBuf) -> None:
        """Remove all objects at the given path.

        Args:
            path (str|Path): The path to the directory.
        """
    def to_async_operator(self) -> AsyncOperator: ...

@final
class AsyncOperator(_Base):
    """The entry class for all public async APIs.

    Args:
        scheme (str): The service name that OpenDAL supports.
        **options (any): The options for the service.
            See the documentation of each service for more details.

    Example:
        ```python
        import opendal
        op = opendal.AsyncOperator("s3", bucket="bucket", region="us-east-1")
        await op.write("hello.txt", b"hello world")
        ```
    """
    def __init__(self, scheme: str, **options: Any) -> None: ...
    def layer(self, layer: Layer) -> "AsyncOperator": ...
    async def open(self, path: PathBuf, mode: str) -> AsyncFile:
        """Open a file at the given path for reading or writing.

        Args:
            path (str|Path): The path to the file.
            mode (str): The mode to open the file. Can be "rb" or "wb".

        Returns:
            A file-like object that can be used to read or write the file.

        Example:
            ```python
            import opendal
            op = opendal.AsyncOperator("s3", bucket="bucket", region="us-east-1")
            async with await op.open("hello.txt", "wb") as f:
                await f.write(b"hello world")
            ```
        """
    async def read(self, path: PathBuf) -> bytes:
        """Read the content of the object at the given path.

        Args:
            path (str|Path): The path to the object.

        Returns:
            The content of the object as bytes.
        """
    async def write(
        self,
        path: PathBuf,
        bs: bytes,
        *,
        append: bool = ...,
        chunk: int = ...,
        content_type: str = ...,
        content_disposition: str = ...,
        cache_control: str = ...,
    ) -> None:
        """Write the content to the object at the given path.

        Args:
            path (str|Path): The path to the object.
            bs (bytes): The content to write.
            append (bool): Whether to append the content to the object.
                Defaults to False.
            chunk (int): The chunk size for writing. Defaults to write all.
            content_type (str): The content type of the object.
                Defaults to None.
            content_disposition (str): The content disposition of the object.
                Defaults to None.
            cache_control (str): The cache control of the object.
                Defaults to None.
        """
    async def stat(self, path: PathBuf) -> Metadata:
        """Get the metadata of the object at the given path.

        Args:
            path (str|Path): The path to the object.

        Returns:
            The metadata of the object.
        """
    async def create_dir(self, path: PathBuf) -> None:
        """Create a directory at the given path.

        Args:
            path (str|Path): The path to the directory.
        """
    async def delete(self, path: PathBuf) -> None:
        """Delete the object at the given path.

        Args:
            path (str|Path): The path to the object.
        """
    async def exists(self, path: PathBuf) -> bool:
        """Check if the object at the given path exists.

        Args:
            path (str|Path): The path to the object.

        Returns:
            True if the object exists, False otherwise.
        """
    async def list(
        self, path: PathBuf, *, start_after: str | None = None
    ) -> AsyncIterable[Entry]:
        """List the objects at the given path.

        Args:
            path (str|Path): The path to the directory.
            start_after (str | None): The key to start listing from.

        Returns:
            An iterable of entries representing the objects in the directory.
        """
    async def scan(self, path: PathBuf) -> AsyncIterable[Entry]:
        """Scan the objects at the given path recursively.

        Args:
            path (str|Path): The path to the directory.

        Returns:
            An iterable of entries representing the objects in the directory.
        """
    async def presign_stat(self, path: PathBuf, expire_second: int) -> PresignedRequest:
        """Generate a presigned URL for stat operation.

        Args:
            path (str|Path): The path to the object.
            expire_second (int): The expiration time in seconds.

        Returns:
            A presigned request object.
        """
    async def presign_read(self, path: PathBuf, expire_second: int) -> PresignedRequest:
        """Generate a presigned URL for read operation.

        Args:
            path (str|Path): The path to the object.
            expire_second (int): The expiration time in seconds.

        Returns:
            A presigned request object.
        """
    async def presign_write(
        self, path: PathBuf, expire_second: int
    ) -> PresignedRequest:
        """Generate a presigned URL for write operation.

        Args:
            path (str|Path): The path to the object.
            expire_second (int): The expiration time in seconds.

        Returns:
            A presigned request object.
        """
    async def presign_delete(
        self, path: PathBuf, expire_second: int
    ) -> PresignedRequest:
        """Generate a presigned URL for delete operation.

        Args:
            path (str|Path): The path to the object.
            expire_second (int): The expiration time in seconds.

        Returns:
            A presigned request object.
        """
    def capability(self) -> Capability: ...
    async def copy(self, source: PathBuf, target: PathBuf) -> None:
        """Copy the object from source to target.

        Args:
            source (str|Path): The source path.
            target (str|Path): The target path.
        """
    async def rename(self, source: PathBuf, target: PathBuf) -> None:
        """Rename the object from source to target.

        Args:
            source (str|Path): The source path.
            target (str|Path): The target path.
        """
    async def remove_all(self, path: PathBuf) -> None:
        """Remove all objects at the given path.

        Args:
            path (str|Path): The path to the directory.
        """
    def to_operator(self) -> Operator: ...

@final
class File:
    """
    A file-like object for reading and writing data.

    Created by the `open` method of the `Operator` class.
    """
    def read(self, size: Optional[int] = None) -> bytes:
        """Read the content of the file.

        Args:
            size (int): The number of bytes to read. If None, read all.

        Returns:
            The content of the file as bytes.
        """
    def readline(self, size: Optional[int] = None) -> bytes:
        """Read a single line from the file.

        Args:
            size (int): The number of bytes to read. If None, read until newline.

        Returns:
            The line read from the file as bytes.
        """
    def write(self, bs: bytes) -> None:
        """Write the content to the file.

        Args:
            bs (bytes): The content to write.
        """
    def seek(self, pos: int, whence: int = 0) -> int:
        """Set the file's current position.

        Args:
            pos (int): The position to set.
            whence (int): The reference point for the position. Can be 0, 1, or 2.

        Returns:
            The new position in the file.
        """
    def tell(self) -> int:
        """Get the current position in the file.

        Returns:
            The current position in the file.
        """
    def close(self) -> None:
        """Close the file."""
    def __enter__(self) -> File:
        """Enter the runtime context related to this object."""
    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        """Exit the runtime context related to this object."""
    @property
    def closed(self) -> bool:
        """Check if the file is closed."""
    def flush(self) -> None:
        """Flush the internal buffer."""
    def readable(self) -> bool:
        """Check if the file is readable."""
    def readinto(self, buffer: Union[bytes, bytearray]) -> int:
        """Read bytes into a buffer.

        Args:
            buffer (bytes|bytearray): The buffer to read into.

        Returns:
            The number of bytes read.
        """
    def seekable(self) -> bool:
        """Check if the file supports seeking."""
    def writable(self) -> bool:
        """Check if the file is writable."""

@final
class AsyncFile:
    """
    A file-like object for reading and writing data.

    Created by the `open` method of the `AsyncOperator` class.
    """
    async def read(self, size: Optional[int] = None) -> bytes:
        """Read the content of the file.

        Args:
            size (int): The number of bytes to read. If None, read all.

        Returns:
            The content of the file as bytes.
        """
    async def write(self, bs: bytes) -> None:
        """Write the content to the file.

        Args:
            bs (bytes): The content to write.
        """
    async def seek(self, pos: int, whence: int = 0) -> int:
        """Set the file's current position.

        Args:
            pos (int): The position to set.
            whence (int): The reference point for the position. Can be 0, 1, or 2.

        Returns:
            The new position in the file.
        """
    async def tell(self) -> int:
        """Get the current position in the file.

        Returns:
            The current position in the file.
        """
    async def close(self) -> None:
        """Close the file."""
    def __aenter__(self) -> AsyncFile:
        """Enter the runtime context related to this object."""
    def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        """Exit the runtime context related to this object."""
    @property
    async def closed(self) -> bool:
        """Check if the file is closed."""
    async def readable(self) -> bool:
        """Check if the file is readable."""
    async def seekable(self) -> bool:
        """Check if the file supports seeking."""
    async def writable(self) -> bool:
        """Check if the file is writable."""

@final
class Entry:
    """An entry in the directory listing."""
    @property
    def path(self) -> str:
        """The path of the entry."""

@final
class Metadata:
    @property
    def content_disposition(self) -> Optional[str]:
        """The content disposition of the object."""
    @property
    def content_length(self) -> int:
        """The content length of the object."""
    @property
    def content_md5(self) -> Optional[str]:
        """The MD5 checksum of the object."""
    @property
    def content_type(self) -> Optional[str]:
        """The mime type of the object."""
    @property
    def etag(self) -> Optional[str]:
        """The ETag of the object."""
    @property
    def mode(self) -> EntryMode:
        """The mode of the object."""

@final
class EntryMode:
    def is_file(self) -> bool: ...
    def is_dir(self) -> bool: ...

@final
class PresignedRequest:
    @property
    def url(self) -> str: ...
    @property
    def method(self) -> str: ...
    @property
    def headers(self) -> dict[str, str]: ...

@final
class Capability:
    """Storage capability information."""

    stat: bool
    """If operator supports stat"""

    stat_with_if_match: bool
    """If operator supports stat with if match"""

    stat_with_if_none_match: bool
    """If operator supports stat with if none match"""

    read: bool
    """If operator supports read"""

    read_with_if_match: bool
    """If operator supports read with if match"""

    read_with_if_none_match: bool
    """If operator supports read with if none match"""

    read_with_override_cache_control: bool
    """If operator supports read with override cache control"""

    read_with_override_content_disposition: bool
    """If operator supports read with override content disposition"""

    read_with_override_content_type: bool
    """If operator supports read with override content type"""

    write: bool
    """If operator supports write"""

    write_can_multi: bool
    """If operator supports write can be called in multi times"""

    write_can_empty: bool
    """If operator supports write with empty content"""

    write_can_append: bool
    """If operator supports write by append"""

    write_with_content_type: bool
    """If operator supports write with content type"""

    write_with_content_disposition: bool
    """If operator supports write with content disposition"""

    write_with_cache_control: bool
    """If operator supports write with cache control"""

    write_multi_max_size: Optional[int]
    """Write_multi_max_size is the max size that services support in write_multi.
    For example, AWS S3 supports 5GiB as max in write_multi."""

    write_multi_min_size: Optional[int]
    """Write_multi_min_size is the min size that services support in write_multi.
    For example, AWS S3 requires at least 5MiB in write_multi expect the last one."""

    write_total_max_size: Optional[int]
    """Write_total_max_size is the max size that services support in write_total.
    For example, Cloudflare D1 supports 1MB as max in write_total."""

    create_dir: bool
    """If operator supports create dir"""

    delete: bool
    """If operator supports delete"""

    copy: bool
    """If operator supports copy"""

    rename: bool
    """If operator supports rename"""

    list: bool
    """If operator supports list"""

    list_with_limit: bool
    """If backend supports list with limit"""

    list_with_start_after: bool
    """If backend supports list with start after"""

    list_with_recursive: bool
    """If backend supports list with recursive"""

    presign: bool
    """If operator supports presign"""

    presign_read: bool
    """If operator supports presign read"""

    presign_stat: bool
    """If operator supports presign stat"""

    presign_write: bool
    """If operator supports presign write"""

    presign_delete: bool
    """If operator supports presign delete"""

    shared: bool
    """If operator supports shared"""

    blocking: bool
    """If operator supports blocking"""
